import sys

from django.core.management import CommandError

from ._private import BaseLiquidbCommand
from ...db_tools import SnapshotCreationHandler, SnapshotHandlerException


class Command(BaseLiquidbCommand):
    help = "Create snapshot of migration state."

    def add_arguments(self, parser):
        parser.add_argument(
            "--name",
            type=str,
            help="Verbose name of snapshot",
        )
        parser.add_argument(
            "--overwrite",
            type=int,
            default=0,
            choices=[0, 1],
            help="If snapshot with given name exist it will overwrite it",
        )

    def _handle(self, *args, **options):
        snapshot_name = options["name"]
        handler = SnapshotCreationHandler(snapshot_name, bool(options["overwrite"]))
        try:
            created = handler.create()
        except SnapshotHandlerException as e:
            raise CommandError(e.error)
        if not created:
            self.stdout.write(
                "All migrations saved in currently applied snapshot. Nothing to create"
            )
            sys.exit(0)
        self.stdout.write(f'Snapshot "{snapshot_name}" successfully save')



import pytest
from liquidb.models import get_latest_applied_migrations_qs

_EXPECTED_APPS = [
    ("first_app", "0001"),
    ("second_app", "0003"),
    ("third_app", "0005"),
]


@pytest.fixture(scope="function")
def _create_apps_fixture(create_migration_state_fixture):
    create_migration_state_fixture(_EXPECTED_APPS)


@pytest.mark.django_db
@pytest.mark.usefixtures("_create_apps_fixture")
def test_latest_applied_migrations_qs():
    ids = get_latest_applied_migrations_qs()
    assert ids.count() == 3
    assert set(ids.values_list("app", flat=True)) == set(
        [app_name for app_name, _ in _EXPECTED_APPS]
    )
    assert set(ids.values_list("name", flat=True)) == set(
        [migration_name for _, migration_name in _EXPECTED_APPS]
    )



import pytest

from liquidb.models import _create_migration_hash


@pytest.mark.parametrize(
    "state_1, state_2,expected",
    [
        (
            [("first_app", "0001"), ("second_app", "0002")],
            [("second_app", "0002"), ("first_app", "0001")],
            True,
        ),
        (
            [],
            [],
            True,
        ),
        (
            [("first_app", "0001"), ("second_app", "0002")],
            [("first_app", "0001"), ("second_app", "0002")],
            True,
        ),
        (
            [("first_app", "0001"), ("second_app", "0002"), ("third_app", "0003")],
            [("first_app", "0001"), ("second_app", "0002")],
            False,
        ),
        (
            [("first_app", "0001"), ("second_app", "0002")],
            [("second_app", "0002")],
            False,
        ),
        (
            [("first_app", "0001"), ("second_app", "0002")],
            [],
            False,
        ),
        (
            [("second_app", "0002")],
            [("first_app", "0001")],
            False,
        ),
    ],
)
def test_hash_function(state_1, state_2, expected):
    s1 = _create_migration_hash(state_1)
    s2 = _create_migration_hash(state_2)
    result = s1 == s2
    assert result is expected



# Generated by Django 3.1.6 on 2021-03-04 08:49

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import liquidb.models
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="MigrationState",
            fields=[
                (
                    "uuid",
                    models.UUIDField(
                        db_index=True,
                        default=uuid.uuid4,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                ("migration_id", models.IntegerField(db_index=True)),
                ("app", models.CharField(max_length=255)),
                ("name", models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name="Snapshot",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.TextField(
                        db_index=True,
                        default=liquidb.models._generate_commit_name,
                        unique=True,
                    ),
                ),
                ("created", models.DateTimeField(default=django.utils.timezone.now)),
                ("applied", models.BooleanField(default=False)),
            ],
        ),
        migrations.AddIndex(
            model_name="snapshot",
            index=models.Index(
                condition=models.Q(applied=True),
                fields=["applied"],
                name="unique_applied",
            ),
        ),
        migrations.AddField(
            model_name="migrationstate",
            name="snapshot",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="migrations",
                to="liquidb.snapshot",
            ),
        ),
    ]



import pytest
from django.core.management import call_command

from liquidb.models import Snapshot


@pytest.mark.django_db
@pytest.mark.parametrize(
    "apps",
    [
        pytest.param(
            [("first_app", "0001"), ("second_app", "0002"), ("third_app", "0003")],
            id="Few apps ",
        ),
        pytest.param(
            [],
            id="Empty migration state",
        ),
    ],
)
def test_delete_snapshot_history(apps, create_migration_state_fixture):
    create_migration_state_fixture(apps)
    call_command("create_migration_snapshot", name="init")
    call_command("delete_snapshot_history", interactive=False)
    assert Snapshot.objects.count() == 0



import os

from django.conf import settings

ADMIN_SNAPSHOT_ACTIONS = os.environ.get("ADMIN_SNAPSHOT_ACTIONS")
if ADMIN_SNAPSHOT_ACTIONS is None:
    ADMIN_SNAPSHOT_ACTIONS = getattr(settings, "ADMIN_SNAPSHOT_ACTIONS", True)



import sys

from django.core.management import CommandError
from django.db import transaction

from ._private import BaseLiquidbRevertCommand


class Command(BaseLiquidbRevertCommand):
    help = "Delete snapshot by name"

    def add_arguments(self, parser):
        parser.add_argument(
            "--name",
            type=str,
            help="Name of snapshot to delete",
        )
        parser.add_argument(
            "--noinput",
            action="store_false",
            dest="interactive",
            default=True,
            help="Do not ask for confirmation ",
        )

    def _handle(self, *args, **options):
        name = options["name"]
        interactive = options["interactive"]
        snapshot = self.get_snapshot(name)
        if interactive:
            confirm = input(
                f"""
                You have requested to delete snapshot "{name}" with all migrations connected to it.
                This action is IRREVERSIBLE.
                Are you sure you want to do this?
                Type 'yes' to continue, or 'no' to cancel:"""
            )
            if confirm != "yes":
                self.stdout.write("Snapshot deletion is canceled")
                sys.exit(0)

        if snapshot.applied:
            raise CommandError(
                f"Snapshot with name {name} is applied and couldn't be deleted"
            )
        with transaction.atomic():
            _total, models = snapshot.delete()
            migrations_states = models.get("liquidb.MigrationState", 0)

        self.stdout.write(
            f'Successfully deleted snapshot "{name}" and {migrations_states} migrations'
        )



from django.core.exceptions import ObjectDoesNotExist
from django.db import transaction, connection
from django.db.migrations.exceptions import NodeNotFoundError
from django.db.migrations.executor import MigrationExecutor
from django.db.migrations.state import ProjectState

from liquidb.models import Snapshot, MigrationState, get_latest_applied_migrations_qs


class SnapshotHandlerException(Exception):
    def __init__(self, error):
        self.error = error
        super().__init__()


class SnapshotCheckoutHandler:
    def __init__(self, snapshot: Snapshot, output=None):
        self.snapshot = snapshot
        self.output = output

    def _write_to_output(self, message):
        """Helper function to wirte to output if specified"""
        if self.output is not None:
            self.output.write(message)

    def _applied_queryset(self):
        return Snapshot.objects.filter(applied=True)

    def _get_lastest_applied_snapshot(self):
        """Return latest applied snapshot by id"""
        try:
            latest = self._applied_queryset().latest("id")
        except ObjectDoesNotExist as error:
            raise SnapshotHandlerException("No latest snapshot present") from error
        return latest

    def _checkout_to_snapshot(self, snapshot: Snapshot):
        """Revert db state to given snapshot"""
        targets = snapshot.migrations.values_list("app", "name")
        if not targets:
            # snapshot that do not have any migrations
            # TODO get all apps from django apps and migrate everything to zero ?
            raise SnapshotHandlerException(
                f"No connected migrations found for snapshot {snapshot.name}"
            )
        executor = MigrationExecutor(connection)
        try:
            _: ProjectState = executor.migrate(targets)
        except KeyError:
            # TODO check exact migration that missing
            message = "\n".join(
                [
                    f"* App Name: {app} ; Migration: {migration}"
                    for app, migration in targets
                ]
            )
            raise SnapshotHandlerException(
                f"\nSome migrations are missing.\n"
                f'Please make sure all migrations in snapshot: "{snapshot.name}";\n'
                f"Are present in corresponding apps: \n{message}"
            )
        except NodeNotFoundError as error:
            raise SnapshotHandlerException(error.message)

    @property
    def applied_snapshot_exists(self):
        return self._applied_queryset().exists()

    def checkout(self, force=False):
        latest = self._get_lastest_applied_snapshot()
        consistent_with_migration_table = latest.consistent_state
        if not consistent_with_migration_table and not force:
            raise SnapshotHandlerException(
                "Migration state is inconsistent latest applied "
                "snapshot do not has all currently applied migrations. "
                "Please use --force flag to checkout anyway."
            )
        if not consistent_with_migration_table:
            self._write_to_output(
                "Force to apply migration. Unsaved changes will be dropped"
            )

        # eq is not same as equal snapshot.id == latest.id
        # see liquidb.models.Snapshot.__eq__
        if self.snapshot == latest:
            # migration hashes are the same
            # nothing to do
            self._write_to_output(f'Snapshot "{latest.name}" already applied ')
            return

        self._checkout_to_snapshot(self.snapshot)
        with transaction.atomic():
            latest.applied = False
            latest.save(update_fields=["applied"])

            self.snapshot.applied = True
            self.snapshot.save(update_fields=["applied"])
        self._write_to_output(
            f'Checkout from snapshot "{latest.name}" to "{self.snapshot.name}"'
        )


class SnapshotCreationHandler:  # pylint: disable=too-few-public-methods
    def __init__(self, snapshot_name, overwrite):
        self.snapshot_name = snapshot_name
        self.overwrite = overwrite
        self.snapshot = None

    def _create(
        self,
        snapshot,
        latest,
    ):
        with transaction.atomic():
            if latest is not None:
                latest.applied = False
                latest.save(update_fields=["applied"])
            migrations = get_latest_applied_migrations_qs(connection).only(
                "id", "app", "name"
            )
            snapshot.applied = True

            if snapshot.id is not None:
                # in overwrite mode
                # so we need to delete all previous migrations
                # from this snapshot
                snapshot.migrations.all().delete()
            snapshot.save()
            to_create = [
                MigrationState(
                    snapshot=snapshot,
                    migration_id=migration.id,
                    app=migration.app,
                    name=migration.name,
                )
                for migration in migrations
            ]
            MigrationState.objects.bulk_create(
                to_create, batch_size=1000, ignore_conflicts=True
            )

    def create(self, dry_run=False) -> bool:
        exists = Snapshot.objects.filter(name=self.snapshot_name).exists()
        if exists and not self.overwrite:
            raise SnapshotHandlerException(
                f"Snapshot with given name {self.snapshot_name} already exists.\n",
            )

        if exists:
            snapshot = Snapshot.objects.get(name=self.snapshot_name)
        else:
            snapshot = Snapshot(name=self.snapshot_name)

        latest = Snapshot.objects.filter(applied=True).last()
        if latest is not None and latest.consistent_state:
            return False
        if dry_run:
            return True

        self._create(snapshot, latest)
        self.snapshot = snapshot
        return True



__all__ = [
    "CheckoutCommand",
    "CheckoutLatestCommand",
    "SnapshotCommand",
]
from ..management.commands.checkout_snapshot import Command as CheckoutCommand
from ..management.commands.checkout_latest_snapshot import (
    Command as CheckoutLatestCommand,
)
from ..management.commands.create_migration_snapshot import Command as SnapshotCommand



from django.conf.urls import url
from django.contrib import admin
from django.contrib.admin import ModelAdmin
from django.core.checks import messages
from django.core.exceptions import ValidationError
from django.forms import ModelForm
from django.http import HttpResponseRedirect
from django.urls import reverse
from django.utils.html import format_html

from .db_tools import (
    SnapshotCheckoutHandler,
    SnapshotHandlerException,
    SnapshotCreationHandler,
)
from .models import Snapshot, MigrationState
from .settings import ADMIN_SNAPSHOT_ACTIONS


class SnapshotAdminModelForm(ModelForm):
    def _create_snapshot(self, name: str, dry_run: bool):  # pylint: disable=no-self-use
        handler = SnapshotCreationHandler(
            name,
            False,
        )
        try:
            created = handler.create(dry_run=dry_run)
        except SnapshotHandlerException as error:
            raise ValidationError(error.error, code="name")
        if not created:
            raise ValidationError(
                "All migrations saved in currently applied snapshot. Nothing to create",
                code="consistent_snapshot",
            )
        return handler.snapshot

    def clean(self):
        cleaned_data = super().clean()
        if self.instance.pk is not None:
            # somehow we modify snapshot
            raise ValidationError(
                "Snapshot is immutable you can't modify it",
                code="immutable_snapshot",
            )
        self._create_snapshot(
            name=self.cleaned_data["name"],
            # Run full validation in SnapshotCreationHandler
            dry_run=True,
        )
        return cleaned_data

    def save(self, commit=True):
        return self._create_snapshot(
            name=self.cleaned_data["name"],
            dry_run=commit,
        )

    def save_m2m(self):  # pylint: disable=method-hidden
        # DO not save any model everything
        # should be done in transaction in
        # self._create_snapshot
        pass


class StateInline(admin.StackedInline):
    model = MigrationState
    can_delete = False
    fk_name = "snapshot"

    # <editor-fold desc="Should be always readonly">
    def has_add_permission(self, request, obj=None):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False

    # </editor-fold>


@admin.register(Snapshot)
class SnapshotAdminView(ModelAdmin):

    form = SnapshotAdminModelForm
    readonly_fields = (
        "created",
        "applied",
        "snapshot_actions",
    )
    list_display = (
        "name",
        "applied",
        "snapshot_actions",
    )
    inlines = [
        StateInline,
    ]

    # <editor-fold desc="Only View Permissions">
    # User should create and delete snapshots through cli
    def has_add_permission(self, request):
        return ADMIN_SNAPSHOT_ACTIONS

    def has_change_permission(self, request, obj=None):
        # never allow edit created snapshot
        return False

    def has_delete_permission(self, request, obj=None):
        return (
            ADMIN_SNAPSHOT_ACTIONS
            and
            # allow delete existing not applied snapshot
            obj is not None
            and not obj.applied
        )

    # </editor-fold>

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = []
        if ADMIN_SNAPSHOT_ACTIONS:
            custom_urls.append(
                url(
                    r"^(?P<snapshot_id>.+)/apply/$",
                    self.admin_site.admin_view(self.apply_snapshot),
                    name="apply_snapshot",
                )
            )
        return custom_urls + urls

    def snapshot_actions(self, obj):
        if ADMIN_SNAPSHOT_ACTIONS and obj.pk is not None and not obj.applied:
            return format_html(
                '<a class="button" href="{}">Apply Snapshot</a>&nbsp;',
                reverse("admin:apply_snapshot", args=[obj.pk]),
            )
        return ""

    snapshot_actions.short_description = "Snapshot Actions"
    snapshot_actions.allow_tags = True

    def apply_snapshot(
        self, request, snapshot_id, *args, **kwargs
    ):  # pylint: disable=unused-argument
        snapshot = self.get_object(request, snapshot_id)
        reversed_url = reverse(
            "admin:liquidb_snapshot_change",
            args=[snapshot.pk],
            current_app=self.admin_site.name,
        )
        handler = SnapshotCheckoutHandler(snapshot)
        if not handler.applied_snapshot_exists:
            self.message_user(
                request,
                "No applied snapshot present",
                level=messages.ERROR,
            )
            return HttpResponseRedirect(reversed_url)

        try:
            # TODO change to apply without force True
            handler.checkout(force=True)
        except SnapshotHandlerException as e:
            self.message_user(
                request,
                e.error,
                level=messages.ERROR,
            )
            return HttpResponseRedirect(reversed_url)

        self.message_user(request, f"Successfully applied {snapshot.name}")
        return HttpResponseRedirect(reversed_url)

    def delete_model(self, request, obj=None):
        # delete snapshot and all related MigrationState
        super().delete_model(request, obj)



from pathlib import Path

SECRET_KEY = "fake-key"

INSTALLED_APPS = [
    # <editor-fold desc="admin part">
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    # </editor-fold>
    "tests",
    "liquidb",
]
# <editor-fold desc="Needed for Admin Part">
BASE_DIR = Path(__file__).resolve().parent.parent

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
]
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [BASE_DIR / "templates"],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]
ROOT_URLCONF = "tests.test_urls"
# </editor-fold>

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3", "NAME": ":memory:"}}
USE_TZ = True



from ._private import BaseLiquidbRevertCommand


class Command(BaseLiquidbRevertCommand):
    help = "Revert migration state to given commit (snapshot)"

    def add_arguments(self, parser):
        parser.add_argument(
            "--name",
            type=str,
            help="Name of snapshot to apply",
        )
        parser.add_argument(
            "--force",
            type=int,
            default=0,
            choices=[0, 1],
            help="Drop unsaved changes",
        )

    def _handle(self, *args, **options):
        snapshot = self.get_snapshot(options["name"])
        self._checkout_snapshot(snapshot, force=bool(options["force"]))



import sys

from django.db import transaction

from ._private import BaseLiquidbRevertCommand
from ...models import Snapshot


class Command(BaseLiquidbRevertCommand):
    help = "Delete all history of snapshots"

    def add_arguments(self, parser):
        parser.add_argument(
            "--noinput",
            action="store_false",
            dest="interactive",
            default=True,
            help="Do not ask for confirmation ",
        )

    def _handle(self, *args, **options):
        interactive = options["interactive"]
        if interactive:
            confirm = input(
                f"""
                You have requested to delete snapshot history with all migrations.
                This action is IRREVERSIBLE.
                Are you sure you want to do this?

                Type 'yes' to continue, or 'no' to cancel: """
            )
            if confirm != "yes":
                self.stdout.write("Snapshot history deletion is canceled")
                sys.exit(0)
        with transaction.atomic():
            _total, models = Snapshot.objects.all().delete()
            snapshots = models.get("liquidb.Snapshot", 0)
            migrations_states = models.get("liquidb.MigrationState", 0)
        self.stdout.write(
            f"Successfully deleted history of {snapshots} snapshots and {migrations_states} migrations"
        )



from django.core.exceptions import ObjectDoesNotExist
from django.core.management import CommandError

from ._private import BaseLiquidbRevertCommand
from ...models import Snapshot


class Command(BaseLiquidbRevertCommand):
    help = "Revert migration to latest commit (snapshot)"

    def add_arguments(self, parser):
        parser.add_argument(
            "--force",
            type=int,
            default=0,
            choices=[0, 1],
            help="Drop unsaved changes",
        )

    def _handle(self, *args, **options):
        try:
            snapshot = Snapshot.objects.latest("id")
        except ObjectDoesNotExist as error:
            raise CommandError("No latest snapshot present") from error
        self._checkout_snapshot(snapshot, force=bool(options["force"]))



import hashlib
from typing import Tuple, List
from uuid import uuid4

from django.db import models, connection
from django.db.migrations.recorder import MigrationRecorder
from django.db.models import Max, Index, Q
from django.utils.crypto import get_random_string
from django.utils.functional import cached_property
from django.utils.timezone import now

from liquidb.consts import SELF_NAME


def _generate_commit_name() -> str:
    """Generates commit name base on datetime"""
    date = now().strftime("%d-%m-%YT%H:%M:%S")
    salt = get_random_string(6)
    return f"{salt}_{date}"


def _create_migration_hash(migration_identifiers: List[Tuple[str, str]]) -> str:
    """create md5 out of tuples of [(app_name, migration_file), ...]"""
    separator = ";"
    inner_separator = "-"
    # always sort alphabetically to ensure same result for unsorted app and sorted
    joined_value = separator.join(
        inner_separator.join([app, name]) for app, name in sorted(migration_identifiers)
    )
    # no reason to check it is only hash of commits
    return hashlib.md5(joined_value.encode("utf-8")).hexdigest()  # nosec  # NOQA


def get_latest_applied_migrations_qs(
    connection_obj=None,
) -> "QuerySet[MigrationRecorder.Migration]":
    """Return latest applied migration in all django apps in project"""
    if connection_obj is None:
        connection_obj = connection
    recorder = MigrationRecorder(connection_obj)
    migration_qs = recorder.migration_qs.exclude(app=SELF_NAME)
    lastest_ids = (
        migration_qs.values("app")
        .annotate(latest_id=Max("id"))
        .values_list(
            "latest_id",
            flat=True,
        )
    )
    return migration_qs.filter(id__in=lastest_ids)


class Snapshot(models.Model):
    name = models.TextField(default=_generate_commit_name, db_index=True, unique=True)
    created = models.DateTimeField(default=now)
    applied = models.BooleanField(default=False)

    class Meta:
        indexes = [
            # by default all applied is false
            # only one row can be applied at the time
            Index(fields=["applied"], name="unique_applied", condition=Q(applied=True))
        ]

    def __repr__(self):
        class_name = self.__class__.__name__
        return f"{class_name}: {self.id}, {self.name}"

    def __eq__(self, other):
        if not isinstance(other, Snapshot):
            return False
        return self.migration_hash == other.migration_hash

    def __hash__(self):
        return hash(self.migration_hash)

    @property
    def consistent_state(self) -> bool:
        """Return True if all connected migrations to current snapshot is applied"""
        current_state = get_latest_applied_migrations_qs().values_list("app", "name")
        current_hash = _create_migration_hash(current_state)
        return self.migration_hash == current_hash

    @cached_property
    def migration_hash(self) -> str:
        """Generate migration hash"""
        return _create_migration_hash(self.migrations.values_list("app", "name"))


class MigrationState(models.Model):
    uuid = models.UUIDField(unique=True, primary_key=True, db_index=True, default=uuid4)
    snapshot = models.ForeignKey(
        Snapshot, on_delete=models.CASCADE, related_name="migrations"
    )
    migration_id = models.IntegerField(db_index=True)
    # redundant info to restore migration
    # from django/db/migrations/recorder.py:30
    app = models.CharField(max_length=255)
    name = models.CharField(max_length=255)

    def __repr__(self):
        return f"Migration {self.app} {self.name}"



from unittest.mock import patch

import pytest
from django.core.management import call_command

from liquidb.db_tools import SnapshotCheckoutHandler
from liquidb.management.commands._private import (
    BaseLiquidbRevertCommand,
)  # NOQA # pylint: disable=protected-access
from liquidb.models import Snapshot
from tests.tools_tests import change_state_mock

_STATE_TEMPLATE = "state_{}"


@pytest.fixture(scope="function")
def _create_three_state_fixture(create_migration_state_fixture):
    state_template = [
        ("first_app", "000{}"),
        ("second_app", "000{}"),
        ("third_app", "000{}"),
    ]
    for counter in range(1, 4):
        state = [
            (app_name, migration.format(counter))
            for app_name, migration in state_template
        ]
        create_migration_state_fixture(state)
        call_command("create_migration_snapshot", name=_STATE_TEMPLATE.format(counter))


@pytest.mark.django_db
@patch.object(
    SnapshotCheckoutHandler,
    "_checkout_to_snapshot",
    new=change_state_mock,
)
def test_checkout_between_snapshots(_create_three_state_fixture):
    state_1 = _STATE_TEMPLATE.format(1)
    state_2 = _STATE_TEMPLATE.format(2)
    state_3 = _STATE_TEMPLATE.format(3)
    call_command("checkout_snapshot", name=state_1)
    snapshot = Snapshot.objects.get(name=state_1)
    assert snapshot.consistent_state is True
    # state_3
    call_command("checkout_snapshot", name=state_3)
    snapshot = Snapshot.objects.get(name=state_3)
    assert snapshot.consistent_state is True
    # state_2
    call_command("checkout_snapshot", name=state_2)
    snapshot = Snapshot.objects.get(name=state_2)
    assert snapshot.consistent_state is True
    # check that previous state not consistent
    snapshot = Snapshot.objects.get(name=state_3)
    assert snapshot.consistent_state is False



from io import StringIO

import pytest
from django.core.management import call_command, CommandError
from mock import patch

from liquidb.models import Snapshot


@pytest.mark.django_db
@pytest.mark.parametrize(
    "apps",
    [
        pytest.param(
            [("first_app", "0001"), ("second_app", "0003"), ("third_app", "0005")],
            id="Three apps",
        ),
        pytest.param([("first_app", "0001"), ("second_app", "0005")], id="Two apps"),
        pytest.param([("first_app", "0001")], id="Single app"),
        pytest.param(
            [],
            id="Empty",
        ),
    ],
)
def test_create_snapshot(apps, create_migration_state_fixture):
    expected = len(apps)
    create_migration_state_fixture(apps)
    s_name = "init"
    call_command("create_migration_snapshot", name=s_name)
    assert Snapshot.objects.count() == 1
    snapshot = Snapshot.objects.first()
    assert snapshot.name == s_name
    assert snapshot.migrations.count() == expected


@pytest.mark.django_db
@pytest.mark.parametrize("overwrite", [0, 1])
def test_snapshot_with_same_name_state_is_not_changed(
    create_migration_state_fixture, overwrite
):
    apps = [("first_app", "0001"), ("second_app", "0005")]
    snapshot_name = "first"
    create_migration_state_fixture(apps)
    call_command("create_migration_snapshot", name=snapshot_name)
    if not overwrite:
        with pytest.raises(CommandError):
            call_command(
                "create_migration_snapshot", name=snapshot_name, overwrite=overwrite
            )
    else:
        with pytest.raises(SystemExit):
            call_command(
                "create_migration_snapshot", name=snapshot_name, overwrite=overwrite
            )


@pytest.mark.django_db
@pytest.mark.parametrize("overwrite", [0, 1])
def test_snapshot_with_same_name_state_is_changed(
    create_migration_state_fixture, overwrite
):
    snapshot_name = "stable"
    state_1 = [("first_app", "0001"), ("second_app", "0005")]
    create_migration_state_fixture(state_1)
    call_command("create_migration_snapshot", name=snapshot_name)
    state_2 = [("first_app", "0002"), ("second_app", "0003"), ("third_app", "0006")]
    create_migration_state_fixture(state_2)
    if not overwrite:
        with pytest.raises(CommandError):
            call_command(
                "create_migration_snapshot", name=snapshot_name, overwrite=overwrite
            )
    else:
        call_command(
            "create_migration_snapshot", name=snapshot_name, overwrite=overwrite
        )
        snapshot = Snapshot.objects.get(name=snapshot_name)
        # check that snapshot overwritten
        assert snapshot.consistent_state is True
        expected_state = state_2
        assert snapshot.migrations.count() == len(expected_state)
        migrations = snapshot.migrations.all()
        assert set(migrations.values_list("app", flat=True)) == set(
            [app_name for app_name, _ in expected_state]
        )
        assert set(migrations.values_list("name", flat=True)) == set(
            [migration_name for _, migration_name in expected_state]
        )


@pytest.mark.django_db
@patch("sys.stdout", new_callable=StringIO)
def test_snapshot_with_same_hash(patched_stdout, create_migration_state_fixture):
    apps = [("first_app", "0001"), ("second_app", "0005")]
    create_migration_state_fixture(apps)
    call_command("create_migration_snapshot", name="first")
    with pytest.raises(SystemExit):
        call_command("create_migration_snapshot", name="second")
    patched_stdout.assert_called_once_with(
        "All migrations saved in currently applied snapshot. Nothing to create"
    )


@pytest.mark.django_db
@patch("sys.stderr")
def test_snapshot_with_same_hash(patched_stderr, create_migration_state_fixture):
    first_state = [("first_app", "0001"), ("second_app", "0005")]
    create_migration_state_fixture(first_state)
    call_command("create_migration_snapshot", name="first")
    second_state = [
        ("first_app", "0003"),
        ("third_app", "0001"),
        ("second_app", "0006"),
    ]
    create_migration_state_fixture(second_state)
    patched_stderr.assert_not_called()



import pytest
from django.core.management import call_command, CommandError

from liquidb.models import Snapshot


@pytest.mark.django_db
def test_delete_snapshot_by_name_applied(create_migration_state_fixture):
    apps = [("first_app", "0001"), ("second_app", "0002"), ("third_app", "0003")]
    create_migration_state_fixture(apps)
    snapshot_name = "init"
    call_command("create_migration_snapshot", name=snapshot_name)
    with pytest.raises(CommandError):
        call_command("delete_snapshot_by_name", name=snapshot_name, interactive=False)
    assert Snapshot.objects.count() == 1


@pytest.mark.django_db
def test_delete_snapshot_by_name_not_applied(create_migration_state_fixture):
    apps = [("first_app", "0001"), ("second_app", "0002"), ("third_app", "0003")]
    create_migration_state_fixture(apps)
    snapshot_name = "first"
    call_command("create_migration_snapshot", name=snapshot_name)
    apps = [("first_app", "0002"), ("second_app", "0003"), ("third_app", "0004")]
    create_migration_state_fixture(apps)
    call_command("create_migration_snapshot", name="second")
    call_command("delete_snapshot_by_name", name=snapshot_name, interactive=False)
    assert Snapshot.objects.count() == 1
    snapshot = Snapshot.objects.first()
    assert snapshot.applied is True
    assert snapshot.consistent_state is True



from unittest.mock import patch

import pytest
from bs4 import BeautifulSoup
from django.urls import reverse

from liquidb.db_tools import SnapshotCheckoutHandler
from liquidb.models import Snapshot
from tests.tools_tests import change_state_mock


def _reverse_snapshot_change(obj):
    return reverse("admin:liquidb_snapshot_change", args=(obj.pk,))


def _inject_managment_form_values(data):
    managment_form_data = {
        "migrations-TOTAL_FORMS": 1,
        "migrations-INITIAL_FORMS": 1,
    }
    data = {**data, **managment_form_data}
    return data


@pytest.fixture(scope="function")
def create_two_snapshots_fixture(create_snapshot_fixture):
    apps_1 = [("first_app", "0001"), ("second_app", "0005")]
    snapshot_one = create_snapshot_fixture(apps_1, "init")
    apps_2 = [("first_app", "0001"), ("second_app", "0006"), ("third_app", "0001")]
    snapshot_two = create_snapshot_fixture(apps_2, "second")
    return snapshot_one, snapshot_two


@pytest.mark.django_db
def test_changeview_snapshot_single(admin_client_fixture, create_snapshot_fixture):
    apps = [("first_app", "0001"), ("second_app", "0005")]
    snapshot = create_snapshot_fixture(apps, "init")
    client = admin_client_fixture
    response = client.get(_reverse_snapshot_change(snapshot))
    assert response.status_code == 200


@pytest.mark.django_db
def test_changeview_two_snapshots_check_not_applied(
    admin_client_fixture, create_two_snapshots_fixture
):
    client = admin_client_fixture
    snapshot_one, snapshot_two = create_two_snapshots_fixture
    snapshot_one.refresh_from_db()
    assert snapshot_one.applied is False
    # not applied snapshot
    response = client.get(_reverse_snapshot_change(snapshot_one))
    assert response.status_code == 200
    soup = BeautifulSoup(response.rendered_content, "html.parser")
    apply_buttons = soup.find_all("div", class_="form-row field-snapshot_actions")
    # assert it is only one button
    assert len(apply_buttons) == 1
    assert apply_buttons[0].a.text == "Apply Snapshot"


@pytest.mark.django_db
def test_changeview_two_snapshots_check_applied(
    admin_client_fixture, create_two_snapshots_fixture
):
    client = admin_client_fixture
    snapshot_one, snapshot_two = create_two_snapshots_fixture
    snapshot_two.refresh_from_db()
    assert snapshot_two.applied is True
    # not applied snapshot
    response = client.get(_reverse_snapshot_change(snapshot_two))
    assert response.status_code == 200
    soup = BeautifulSoup(response.rendered_content, "html.parser")
    apply_buttons = soup.find_all("div", class_="form-row field-snapshot_actions")
    # assert it is only one button
    assert len(apply_buttons) == 1
    # this snapshot already applied
    assert apply_buttons[0].a is None


@pytest.mark.django_db
def test_changelist_snapshot_empty(admin_client_fixture):
    client = admin_client_fixture
    url = reverse("admin:liquidb_snapshot_changelist")
    response = client.get(url)
    assert response.status_code == 200


@pytest.mark.django_db
def test_changelist_snapshot_not_empty(
    admin_client_fixture, create_migration_state_fixture, create_snapshot_fixture
):
    client = admin_client_fixture
    apps = [("first_app", "0001"), ("second_app", "0005")]
    _snapshot = create_snapshot_fixture(apps, "init")
    url = reverse("admin:liquidb_snapshot_changelist")
    response = client.get(url)
    assert response.status_code == 200

    soup = BeautifulSoup(response.rendered_content, "html.parser")
    assert len(soup.find_all("a", class_="addlink")) == 1


@pytest.mark.django_db
def test_changelist_snapshot_two_snapshots(
    admin_client_fixture, create_migration_state_fixture, create_two_snapshots_fixture
):
    client = admin_client_fixture
    snapshot_one, snapshot_two = create_two_snapshots_fixture
    url = reverse("admin:liquidb_snapshot_changelist")
    response = client.get(url)
    assert response.status_code == 200

    soup = BeautifulSoup(response.rendered_content, "html.parser")
    change_list = soup.find(id="result_list")
    assert len(change_list.find_all("th", class_="field-name")) == 2
    assert str(snapshot_two.id) in change_list.find_all("th", class_="field-name")[
        0
    ].a.get("href")
    assert str(snapshot_one.id) in change_list.find_all("th", class_="field-name")[
        1
    ].a.get("href")
    snapshot_one.refresh_from_db()
    assert snapshot_one.applied is False
    apply_buttons = soup.find_all("td", class_="field-snapshot_actions")
    # second is empty because it is already applied
    assert apply_buttons[0].a is None
    # get apply button for snapshot one
    apply_button = apply_buttons[-1]
    assert str(snapshot_one.id) in apply_button.a.get("href")


@pytest.mark.django_db
def test_create_init_snapshot(admin_client_fixture, create_migration_state_fixture):
    name = "init"
    create_migration_state_fixture([("third_app", "0001"), ("five_app", "0005")])
    client = admin_client_fixture
    response = client.post(
        reverse("admin:liquidb_snapshot_add"),
        _inject_managment_form_values({"name": name}),
        follow=True,
    )
    assert response.status_code == 200
    snapshot = Snapshot.objects.get(name=name)
    assert snapshot.applied is True
    assert snapshot.consistent_state is True


@pytest.mark.django_db
def test_create_snapshot_with_existing(
    admin_client_fixture, create_snapshot_fixture, create_migration_state_fixture
):
    create_snapshot_fixture([("third_app", "0002"), ("five_app", "0005")], "init")
    client = admin_client_fixture
    create_migration_state_fixture([("third_app", "0003"), ("five_app", "0006")])
    snapshot_name = "following"
    response = client.post(
        reverse("admin:liquidb_snapshot_add"),
        _inject_managment_form_values({"name": snapshot_name}),
        follow=True,
    )
    assert response.status_code == 200
    snapshot = Snapshot.objects.get(name=snapshot_name)
    assert snapshot.applied is True
    assert snapshot.consistent_state is True


@pytest.mark.django_db
def test_existing_snapshot(admin_client_fixture, create_snapshot_fixture):
    name = "init"
    create_snapshot_fixture([("third_app", "0001"), ("five_app", "0005")], name)
    client = admin_client_fixture
    response = client.post(
        reverse("admin:liquidb_snapshot_add"),
        _inject_managment_form_values({"name": name}),
        follow=True,
    )
    soup = BeautifulSoup(response.rendered_content, "html.parser")
    errors = soup.find("ul", class_="errorlist nonfield")
    assert f"Snapshot with given name {name} already exists" in errors.text


@pytest.mark.django_db
def test_consistent_state_snapshot(admin_client_fixture, create_snapshot_fixture):
    create_snapshot_fixture([("third_app", "0002"), ("five_app", "0005")], "init")
    client = admin_client_fixture
    response = client.post(
        reverse("admin:liquidb_snapshot_add"),
        _inject_managment_form_values({"name": "second"}),
        follow=True,
    )
    soup = BeautifulSoup(response.rendered_content, "html.parser")
    errors = soup.find("ul", class_="errorlist nonfield")
    assert (
        "All migrations saved in currently applied snapshot. Nothing to create"
        in errors.text
    )


@pytest.mark.django_db
@patch.object(
    SnapshotCheckoutHandler,
    "_checkout_to_snapshot",
    new=change_state_mock,
)
def test_apply_snapshot(
    admin_client_fixture, create_two_snapshots_fixture, create_migration_state_fixture
):
    snapshot_1, snapshot_2 = create_two_snapshots_fixture
    client = admin_client_fixture
    snapshot_1.refresh_from_db()
    assert snapshot_1.applied is False
    response = client.post(
        reverse("admin:apply_snapshot", args=(snapshot_1.pk,)), follow=True
    )
    assert response.status_code == 200
    # check that first snapshot is applied
    snapshot_1.refresh_from_db()
    assert snapshot_1.consistent_state is True
    assert snapshot_1.applied is True
    # check that second snapshot is not applied
    snapshot_2.refresh_from_db()
    assert snapshot_2.applied is False
    assert snapshot_2.consistent_state is False



from abc import ABCMeta, abstractmethod

from django.core.exceptions import ObjectDoesNotExist
from django.core.management import BaseCommand, CommandError
from django.db import connection
from django.db.migrations.recorder import MigrationRecorder

from ...db_tools import SnapshotCheckoutHandler, SnapshotHandlerException
from ...models import Snapshot, MigrationState, get_latest_applied_migrations_qs


class BaseLiquidbCommand(BaseCommand, metaclass=ABCMeta):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.connection = connection

    def _check_migration_db_exists(self):
        recorder = MigrationRecorder(self.connection)
        if not recorder.has_table():
            raise CommandError("No migration table present")

    def _check_liquidb_tables_exists(self):
        with self.connection.cursor() as cursor:
            tables = self.connection.introspection.table_names(cursor)
        tables_to_check = [
            Snapshot._meta.db_table,
            MigrationState._meta.db_table,
        ]  # NOQA
        if tables and tables_to_check:
            return all(map(lambda x: x in tables, tables_to_check))
        return False

    def _init(self):
        self._check_migration_db_exists()
        self._check_liquidb_tables_exists()

    def handle(self, *args, **options):
        self._init()
        self._handle(*args, **options)

    def _latest_migrations(self):
        return get_latest_applied_migrations_qs(self.connection)

    @staticmethod
    def get_snapshot(name: str) -> Snapshot:
        """Retrun snapshot by name"""
        try:
            snapshot = Snapshot.objects.get(name=name)
        except ObjectDoesNotExist as error:
            raise CommandError(
                f'Snapshot with name: "{name}" doesn\'t exists'
            ) from error
        return snapshot

    @abstractmethod
    def _handle(self, *args, **options):
        raise NotImplementedError


class BaseLiquidbRevertCommand(BaseLiquidbCommand, metaclass=ABCMeta):
    def _checkout_snapshot(self, snapshot: Snapshot, force=False):
        """Run all checks before reverting to desired state"""

        handler = SnapshotCheckoutHandler(snapshot, self.stdout)
        if not handler.applied_snapshot_exists:
            raise CommandError("No latest snapshot present")

        try:
            handler.checkout(force=force)
        except SnapshotHandlerException as e:
            raise CommandError(e.error)

    @abstractmethod
    def _handle(self, *args, **options):
        raise NotImplementedError
